#!/usr/bin/env python3

import rospy
import smach
import smach_ros
import tf2_ros
import actionlib

import math
import json

from move_base_msgs.msg import MoveBaseAction, MoveBaseGoal
from std_msgs.msg import String, Bool
from unsw_action_msg.msg import UNSWActionMsg, UNSWActionResult

class WaitForAction(smach.State) :
    def __init__(self) :
        smach.State.__init__(self, outcomes=['goto_goal_action', 'waiting'], input_keys=['data'], output_keys=['data'])
        action_subcriber = rospy.Subscriber("/actions", UNSWActionMsg , self.action_cb, queue_size=1)
        self.action = None
        self.data = None

    def action_cb(self, msg: UNSWActionMsg):
        self.data = json.loads(f"{msg.data}")
        if msg.action_name == "goto_goal" :
            self.action = "goto_goal_action"
        elif msg.action_name == "follow_person":
            self.action = "follow_person_action"

    def execute(self, ud):
        ud.data = self.data
        if self.action is not None:
            action_to_return = self.action
            self.action = None  # Reset the action to prevent re-triggering
            return action_to_return
        
        return 'waiting'  # Default outcome if no action is set


# '{"GoalPose":
#     {
#         "header":{"frame_id" : "map"},
#         "pose": {
#             "position" : {
#                 "x": -4.137215614318848,
#                 "y": -0.4085843563079834,
#                 "z": 0
#             },
#             "orientation" : {
#                 "x" : 0.0,
#                 "y" : 0.0,
#                 "z" : -0.3196001341628789,
#                 "w" : 0.9475525073804986
#             }
#         }
#     }
# }'

# '{"GoalPose":{"header":{"frame_id" : "map"},"pose": {"position" : {"x": -4.137215614318848,"y": -0.4085843563079834,"z": 0},"orientation" : {"x" : 0.0,"y" : 0.0,"z" : -0.3196001341628789,"w" : 0.9475525073804986}}}}'

class GotoGoalAction(smach.State):
    def __init__(self) :
        smach.State.__init__(self, outcomes=['success', 'failure'], input_keys=['data'])
        self.client = actionlib.SimpleActionClient('move_base', MoveBaseAction)
        if not self.client.wait_for_server(rospy.Duration(5.0)):
            rospy.logerr("Timeout waiting for /move_base/move action server")
            assert False, "Timeout waiting for action server"

    def execute(self, ud):

        goal = MoveBaseGoal()
        goal.target_pose.header.stamp = rospy.Time.now()
        goal.target_pose.header.frame_id = ud.data['GoalPose']['header']['frame_id']
        goal.target_pose.pose.position.x = ud.data['GoalPose']['pose']['position']['x']
        goal.target_pose.pose.position.y = ud.data['GoalPose']['pose']['position']['y']

        goal.target_pose.pose.orientation.x = ud.data['GoalPose']['pose']['orientation']['x']
        goal.target_pose.pose.orientation.y = ud.data['GoalPose']['pose']['orientation']['y']
        goal.target_pose.pose.orientation.z = ud.data['GoalPose']['pose']['orientation']['z']
        goal.target_pose.pose.orientation.w = ud.data['GoalPose']['pose']['orientation']['w']
        self.client.send_goal(goal)

        result = self.client.wait_for_result()
        if result:
            return 'success'
        else:
            return 'failure'
    
class GotoGoalActionSucces(smach.State):
    def __init__(self):
        smach.State.__init__(self,outcomes=['success'])
        self.results_pub = rospy.Publisher("actions/result", UNSWActionResult, queue_size=1)
    
    def execute(self, ud):
        result = UNSWActionResult()
        result.action_name = "goto_goal"
        result.action_result = True
        self.results_pub.publish(result)
        return 'success'
    
class GotoGoalActionFailure(smach.State):
    def __init__(self):
        smach.State.__init__(self,outcomes=['failure'])
        self.results_pub = rospy.Publisher("actions/result", UNSWActionResult, queue_size=1)
    
    def execute(self, ud):
        result = UNSWActionResult()
        result.action_name = "goto_goal"
        result.action_result = False
        self.results_pub.publish(result)
        return 'failure'

def main():
    rospy.init_node('navigation_client_smach')
    sm = smach.StateMachine(outcomes = ["action_success"])

    with sm:
        smach.StateMachine.add('WAIT_FOR_ACTION', 
                               WaitForAction(), 
                               transitions={'goto_goal_action':'GOTO_GOAL_ACTION',
                                            'waiting' : 'WAIT_FOR_ACTION'
                                            })
        smach.StateMachine.add('GOTO_GOAL_ACTION',
                               GotoGoalAction(),
                               transitions={
                                   'success':'GOTO_GOAL_ACTION_SUCCESS',
                                   'failure' : 'GOTO_GOAL_ACTION_FAILURE'
                               })
        smach.StateMachine.add('GOTO_GOAL_ACTION_SUCCESS',
                               GotoGoalActionSucces(),
                               transitions={
                                   'success' : 'WAIT_FOR_ACTION'
                               })
        smach.StateMachine.add('GOTO_GOAL_ACTION_FAILURE',
                               GotoGoalActionFailure(),
                               transitions={
                                   'failure' : 'WAIT_FOR_ACTION'
                               })

    sis = smach_ros.IntrospectionServer('server_name',sm,'/SM_ROOT')
    sis.start()

    outcome = sm.execute()
    rospy.logwarn(f"State machine finished with outcome: {outcome}")
    rospy.spin()
    sis.stop()

if __name__== "__main__":
    main()



