#! /usr/bin/env python3

import rospy
import actionlib
import hsrb_interface
from hsrb_interface import geometry
import tf2_geometry_msgs
from action_server.msg import ManipApproachAction, ManipApproachFeedback, ManipApproachResult
from math import pi
import tf2_ros

import math
from move_base_msgs.msg import MoveBaseAction, MoveBaseGoal

from tf.transformations import quaternion_from_euler


PALM_SIZE = 0.06
class ManipApproach(object):
    def __init__(self, name):
        # init the robot
        self._robot = hsrb_interface.Robot()
        self._whole_body = self._robot.get('whole_body')
        self._gripper = self._robot.get('gripper')

        self.move_base_action_client = actionlib.SimpleActionClient('move_base', MoveBaseAction)
        if not self.move_base_action_client.wait_for_server(rospy.Duration(10)):
            rospy.logerr("move_base action server is not available")
            rospy.signal_shutdown("move_base action server is not available")
            return
        self.tfBuffer = tf2_ros.Buffer()
        self.listener = tf2_ros.TransformListener(self.tfBuffer)
        self._action_name = name
        self._as = actionlib.SimpleActionServer(self._action_name, ManipApproachAction, execute_cb=self.execute_cb, auto_start = False)
        self._as.start()
        rospy.loginfo("Action server manip-approach started")


    # open the gripper, can be extracted to a helper function of robot class
    def open_gripper(self):
        rospy.loginfo("opening gripper")
        self._gripper.command(1.2)
    # close the gripper, can be extracted to a helper function of robot class
    def close_gripper(self):
        self._gripper.apply_force(0.7)

    # move the arm lift joint to allign with the object height to avoid collision effectively
    def arm_joint_init(self, target_height):
        # when joint value is 0, it is 33cm from actual ground.
        joint_zero_rel_base = 0.33
        joint_values = {'arm_lift_joint': target_height - joint_zero_rel_base}
        self._whole_body.move_to_joint_positions(joint_values)

    def execute_cb(self, goal):
        r = rospy.Rate(1)
        feedback = ManipApproachFeedback()
        result = ManipApproachResult()
        # default to be true
        result.action_success=True

        feedback.feedback = 0  # Not started
        self._as.publish_feedback(feedback)

        direction = goal.direction
        target_point = goal.target_point
        self._whole_body.move_to_neutral()

        feedback.feedback = 1  # Ongoing
        self._as.publish_feedback(feedback)

        try:
            trans = self.tfBuffer.lookup_transform("base_footprint", target_point.header.frame_id, rospy.Time.now(), rospy.Duration(2))
            transformed_point_stamped = tf2_geometry_msgs.do_transform_point(target_point, trans)
            angle_to_target = math.atan2(transformed_point_stamped.point.y, transformed_point_stamped.point.x)
            # print(angle_to_target)
            # rotate the robot to face the target and then another 90 degree make the right side of the robot facing the target
            # calc the new posestamp in base_footprint frame(0 + angle_to_target)
            goal_pose_rotation = tf2_geometry_msgs.PoseStamped()
            goal_pose_rotation.header.frame_id = "base_footprint"
            # calc the quanternion from ei ej ek use euler to quaternion
            q = quaternion_from_euler(0,0,angle_to_target)
            # print(q)
            goal_pose_rotation.pose.orientation.w = q[3]
            goal_pose_rotation.pose.orientation.z = q[2]
            goal_pose_rotation.pose.orientation.y = q[1]
            goal_pose_rotation.pose.orientation.x = q[0]
            trans2 = self.tfBuffer.lookup_transform("map", "base_footprint", rospy.Time.now(), rospy.Duration(2))
            transformed_pose_stamp_map = tf2_geometry_msgs.do_transform_pose(goal_pose_rotation, trans2)
            
            goal = MoveBaseGoal()
            goal.target_pose = transformed_pose_stamp_map
            self.move_base_action_client.send_goal_and_wait(goal)
        except Exception as e:
            rospy.loginfo("Failed to rotate to the object. {}".format(e))
            feedback.feedback = 2  # Failure
            self._as.publish_feedback(feedback)
            result.action_success = False
            result.forward_distance = 0
            self._as.set_aborted(result)
            return
        
        rospy.sleep(2)
        try:
            trans = self.tfBuffer.lookup_transform("base_footprint", target_point.header.frame_id, rospy.Time.now(), rospy.Duration(2))
            transformed_point_stamped = tf2_geometry_msgs.do_transform_point(target_point, trans)
            rospy.loginfo(transformed_point_stamped)
        except Exception as e:
            rospy.loginfo("Failed to transform point to footprint. {}".format(e))
            feedback.feedback = 2  # Failure
            self._as.publish_feedback(feedback)
            result.action_success = False
            result.forward_distance = 0
            self._as.set_aborted(result)
            return
        
        
        line_dis = 0
        if direction == "top":
            pose = geometry.pose(x=transformed_point_stamped.point.x,
                            y=transformed_point_stamped.point.y,
                            z=min(max(float(transformed_point_stamped.point.z+ 0.1),0.02),1.00),
                            ei=-pi,
                            ej=0,
                            ek=-pi/2)
            line_dis =  pose.pos.z - transformed_point_stamped.point.z - PALM_SIZE

        elif direction == "top2":
            pose = geometry.pose(x=transformed_point_stamped.point.x,
                            y=transformed_point_stamped.point.y,
                            z=min(max(float(transformed_point_stamped.point.z+ 0.1),0.02),1.00),
                            ei=-pi,
                            ej=0,
                            ek=0)
            line_dis =  pose.pos.z - transformed_point_stamped.point.z - PALM_SIZE
        elif direction == "front":
            pose = geometry.pose(x=transformed_point_stamped.point.x-0.3,
                                y=transformed_point_stamped.point.y,
                                z=min(max(float(transformed_point_stamped.point.z),0.02),2.00),
                                ei=-pi,
                                ej=-pi/2,
                                ek=0)
            line_dis =  transformed_point_stamped.point.x - pose.pos.x - PALM_SIZE
        elif direction == "front-vertical":
            pose = geometry.pose(x=transformed_point_stamped.point.x-0.3,
                                y=transformed_point_stamped.point.y,
                                z=min(max(float(transformed_point_stamped.point.z),0.02),1.00),
                                ei=-pi/2,
                                ej=0,
                                ek=-pi/2)
            line_dis =  transformed_point_stamped.point.x - pose.pos.x - PALM_SIZE
        else:
            rospy.loginfo("Invalid direction")
            feedback.feedback = 2
            self._as.publish_feedback(feedback)
            result.action_success = False
            result.forward_distance = 0
            self._as.set_aborted(result)
            return
        # acually move the robot
        try:
            self.open_gripper()
            # move to that pose
            self.arm_joint_init(pose.pos.z)
            self._whole_body.end_effector_frame = u'hand_palm_link'
            self._whole_body.move_end_effector_pose(pose,
                                                    "base_footprint")
            
            # move it to allign better
            # To overcome the tf misallign for hand plam link
            # THIS IS A HACK NOW NEED TO BE REVISED LATER
            if direction in ["top", "top2"]:
                line_traj = (-1, 0, 0)
                l =  0.03
                self._whole_body.move_end_effector_by_line(line_traj,l)
            # return the value if success, otherwise the error will be captured and false will be returned
            result.action_success = True
            result.forward_distance = line_dis
            self._as.set_succeeded(result)

            # set the feedback to not started
            feedback.feedback = 0  # Not started
            self._as.publish_feedback(feedback)
            return
        except Exception as e:
            rospy.logwarn("Failed moving body. {}".format(e))
            self._as.set_aborted(result)
            return

        


        
        


if __name__ == '__main__':
    rospy.init_node('manip_approach')
    server = ManipApproach(rospy.get_name())
    rospy.spin()
    