#!/usr/bin/env python3

# Person Recoginition 
# The person recognition node is used to capture and re-identify people in an image. 
# The image can be a video stream frame.The idea is to be able to create a node where components 
# can be added and removed in a modular way
#
# Person recognition is made up of a number components:
# Face Recognition - recognising the specific features on a human face
# Feature recognition - capturing different features e.g. shirt texture
#
# This code can be run with or without the World Model
#
# Libraries:
# face recognition - https://github.com/ageitgey/face_recognition

# Dependencies:
# World Model - database with schema 2.1.14 or higher which contains the agent table structures - use with flag: use_world_model = True

# Written By: Adam Golding

from datetime import datetime
import rospy
import sys
import os
from pathlib import Path
# from skimage import feature
# from scipy.spatial import distance
from math import sqrt
import cv2
import numpy as np
from sklearn.metrics.pairwise import euclidean_distances
from sklearn.preprocessing import MinMaxScaler
from sklearn.neighbors import NearestNeighbors
from tensorflow.keras.applications import VGG19
from tensorflow.keras.models import Model
from tensorflow.keras.layers import GlobalAveragePooling2D
from keras.applications.vgg19 import preprocess_input
import tensorflow as tf
# from visualizer import draw_face_bbox, draw_face_name, draw_face_info_from_msg, draw_detections, plot_skeleton_kpts, print_objects, display_image, visualize_image, print_agent, print_known_agents
from visualizer import display_image, visualize_image, print_agent, print_known_agents
module_path = os.environ.get("UNSW_WS")
sys.path.append(module_path + "/WORLD_MODEL/world_model_kb/scripts")
from unsw_vision_msgs.msg import BoundingBox
from agent_api import AgentApi
from dbase import DBase
from person import Person
from feature import Feature
import face_recognition


class PersonRecognition():
    def __init__(self, db_ini, scale, feature_threshold, face_threshold, use_world_model, save_img, img_path) -> None:
        self.db_ini = db_ini
        self.scale = scale
        self.feature_threshold = feature_threshold
        self.face_threshold = face_threshold
        self.use_world_model = use_world_model
        self.save_img = save_img
        self.img_path = img_path
        self.feature_type_list = ["face", "head"]
        # self.feature_type_list = ["face", "shirt", "pants", "head"]
        self.feature_attr_list = ["encoding"]
        self.perspective_list = ["front", "front-left", "back-left", "back", "back-right", "front-right"]
        vgg_gpu_options = tf.compat.v1.GPUOptions(per_process_gpu_memory_fraction=0.1)  # Allocate 10% of GPU memory for VGG model
        vgg_config = tf.compat.v1.ConfigProto(gpu_options=vgg_gpu_options)
        vgg_session = tf.compat.v1.Session(config=vgg_config)
        self.agent = AgentApi()
        self.db = None
        self.crop_w = 75
        self.crop_h = 120  #135
        self.h_multiplier = 1.6   # run01 = 1.3, run02 = 1.5; run03 = 1.8; run04 = 2.0; run05 = 2.5; run08 = 1.5
        self.feature_extractor = self.init_feature_extractor()
        # print(f"crop_w = {self.crop_w}, crop_h = {self.crop_h} multipler = {self.h_multiplier}")

        self.known_agents = []     # list of known Persons

        np.set_printoptions(threshold=np.inf)
        # if world model used load agent data from database
        if self.use_world_model:
            self.db = DBase(os.path.abspath(self.db_ini))               # database connections           
            self.known_agents = self.agent.get_agent_list(self.db.con_pool)
            self.print_known_agents()
        rospy.loginfo(f"Number of existing agents loaded from the world model = {len(self.known_agents)}")
    

    def init_feature_extractor(self):
        # Define the base CNN model (e.g., ResNet50)
        # base_model = MobileNetV3Small(
        #     input_shape=(self.crop_h, self.crop_w, 3),
        #     include_top=False,
        #     weights='imagenet'
        # )
        # base_model = MobileNetV2(
        #     input_shape=(self.crop_h, self.crop_w, 3),
        #     include_top=False,
        #     weights='imagenet'
        # )        
        base_model = VGG19(
            input_shape=(self.crop_h, self.crop_w, 3),
            include_top=False,
            weights='imagenet'
        )                 
        global_avg_pooling_layer = GlobalAveragePooling2D()(base_model.output)
        feature_extractor_model = Model(inputs=base_model.input, outputs=global_avg_pooling_layer)
        return feature_extractor_model        
    
    '''
    Function to extract person and features from a detected person in an image. 
    The person object with features is returned for matching against known people.
    The intention is that this function can be expanded to include additional feature extraction processes.

    Parameters:
    img     - the original image
    img_person  - the person of interest in the original image
    face_locations  - list of face locations within the image
    face_encodings  - list of face encoding data for faces within the image

    Return:
    Person object containing details and features that have been extracted from the person of interest
    '''
    def extract_person_features(self, img, person_msg, face_locations, face_encodings):
        try:
            _person = Person(None, person_msg.bbox_person, person_msg.bbox_head, person_msg.bbox_face, person_msg.skeleton, person_msg.position, None)  

            # get the perspective of the person so we know how the robot sees them
            # (front, front-left, back-left, back, back-right, front-right)
            perspective = self.get_perspective(person_msg.skeleton)
            print(f"Perspective = {perspective}")
            # if we detect the perspective we can try to extract features 
            if perspective is not None:                      
                for feature_type in self.feature_type_list: 
                    crop_img = None # reset each iteration
                    encoding = None
                    if feature_type == "face":
                        if len(face_encodings) > 0:                         
                            # try to align the face with the skeleton 
                            _face_encoding = self.align_face_to_head(person_msg, face_locations, face_encodings)                  
                            new_feat = Feature(perspective, feature_type, _face_encoding)
                            _person.add_feature(new_feat)
                    else:                  
                        print(f"feture type = {feature_type}")      
                        # crop the region of interest
                        if feature_type == "shirt":                    
                            crop_img = self.crop_shirt(img, person_msg.skeleton, perspective)   
                        elif feature_type == "pants":
                            pass
                        elif feature_type == "head":
                            crop_img = self.crop_roi(img, _person.bbox_head, perspective) 

                        # extract the feature
                        if crop_img is not None:
                            encoding = self.get_feature(crop_img)   
                            new_feat = Feature(perspective, feature_type, encoding)                                                
                            _person.add_feature(new_feat)
                        else:
                            print("Crop is none")
                            _person = None
            else:
                _person = None
            
            return _person
        except Exception as e:
            rospy.logwarn(f"Extract Person Features Error: {e} ")
            # print("Extract Person Features Error: ",e)
            return None

    '''
    Align the face with the head of the person (using face bbox and skeletal keypoints)
    use keypoints 0,1,2 = 0 (nose), 1 (right eye) and 2 (left eye)     
    For the person of interest we search all existing faces
    The face bbox is generated by the DLIB library which uses a diferent format to the BoundingBox msg
    The face bbox needs to be converted to the BoundingBox msg format
    Check to see if ANY of the 3 keypoints exist inside the face bounding box
    i.e. 0 or 1 or 2
    
    Parameters:
    bbox = bounding box of the face
    person = the person from the message which includes skeletal keypoints

    Returns:
    Tuple - matched bounding box of face, face encoding
    '''
    def align_face_to_head(self, person, face_locations, face_encodings):      
        for (top, right, bottom, left), face_encoding in zip(face_locations, face_encodings):  
            # convert dlib bbox coordinates to BoundingBox msg format
            x = left
            y = top
            width = right-left
            height = bottom-top
            # rescale (to the original image size) the bounding box coordinates for the face
            bbox = self.rescale_bbox(x, y, width, height, self.scale)            

            x1 = bbox.x 
            y1 = bbox.y
            x2 = bbox.x + bbox.width        
            y2 = bbox.y + bbox.height
            inside = False   #initialise

            # check to see if keypoint are within the bounding box of the face
            i=0
            while i < 3 and inside == False:
                if person.skeleton[i].x != 0.0 and person.skeleton[i].y != 0.0:                    
                    if int(person.skeleton[i].x) > int(x1) \
                        and int(person.skeleton[i].x) < int(x2) \
                        and int(person.skeleton[i].y) > int(y1) \
                        and int(person.skeleton[i].y) < int(y2): 

                        inside = True

                i += 1
            
            if inside:
                return face_encoding

        return None

    '''
    Function to compare all features of a person against known people/agents.

    If we can positively match the face we don't need to match against any other features

    Parameters:
    person - the detected person which we are trying to match to known people

    Return:
    The matched person
    '''
    def match_person(self, person):
        id = None
        # check the face features first as this gives a more confident match
        for feat in person.features:
            if feat.feature_type == "face":
                id = self.match_face(feat.encoding)
                
                if id is not None:
                    person.id = id      # update the id of the person object
                    print(f"Found a match based on face recognition {id}")
                    return person    # facial features were extracted               

        # if the face was not matched then we need to check the other features
        for feat in person.features:
            if feat.feature_type != "face":
                agent_feature_list = self.get_features_for_known_agents(person, 1)
                id = self.match_features(feat.encoding, agent_feature_list)              

        if id is not None:
            person.id = id      # update the id of the person object
            print(f"Found a match based on feature recognition: ID = {id}")

        return person

    '''
    Compares the new face to that in existing/known people

    Parameters:
    face_encoding = the new face_encoding we are comparing to known face encodings

    Return:
    if matched found then the id of the matched agent/person
    if no match then None
    '''
    def match_face(self, face_encoding):
        face_distances, agent_list = self.face_distance(face_encoding)
        print(f"Face Distances: {face_distances}")
        print(f"Face Agent: {agent_list}")

        if face_distances is not None: 
            if len(face_distances) > 0:    
                min_idx = np.argmin(face_distances)
                if face_distances[min_idx] <= self.face_threshold:    # found a match                                          
                    _person = self.known_agents[agent_list[min_idx]]
                    return _person.id
                else:
                    return None
            else:
                return None
        else:
            return None       

    '''
    Not Used???
    This function will build a single feature array based on the available features in the detected person.

    Parameters:
    person = the detected person

    Return:
    a list of values encoding for each feature
    '''
    def get_features_for_person(self, person):
        feature_list = []
        feature_type_list = []

        for feat in person.features:
            if (feat.feature_type != "face"
                and feat.encoding is not None): 
                    # each feature is added to the list
                    # e.g. (0.123, 0.456, ...., 0.543)                    
                    feature_list.extend(feat.encoding)
                    feature_type_list.append(feat.feature_type) 

        print(f"Detected Person Feature Types: {feature_type_list}")                          
        # print(f"Detected Person Feature Types: {feature_type_list} {feature_list}")
        num_features = len(feature_type_list)
        return feature_list, num_features

    '''
    This function will build a single feature array based on the available features in the detected person

    Parameters:
    person = the detected person
    agent = the known agent

    Return:
    a list of dictionary {id; concatenated features} for each known agent
    '''
    def get_features_for_known_agents(self, person, num_features):
        vector_list = []     # store the list of tuples (agent_id, feature_vector)

        for agent in self.known_agents:  
            for perspective in self.perspective_list:            
                for person_feature in person.features:
                    if person_feature.feature_type != "face":
                        encoding = agent.get_feature(perspective, person_feature.feature_type) # contains both feature encoding
                        if encoding is not None:
                            vector_list.append((agent.id, encoding))  
                  
        return vector_list

    '''
    Compares the texture of the new information to that in existing/known people

    Parameters:
    person = contains the features, texture that are being compared to known agents
    feature_type = the item being compared, e.g. shirt, pants etc

    Return:
    if matched found then the id of the matched agent/person
    if no match then None
    '''
    def match_features(self, person_features, agent_feature_list):        
        if len(agent_feature_list) > 0:            
            # Separate the IDs and the feature vectors
            agent_ids, features = zip(*agent_feature_list)
            # Convert features to a 2D array
            features = np.vstack(features)
            # Fit a NearestNeighbors model on agent_list_features
            neighbors = NearestNeighbors(n_neighbors=1).fit(features)
            # Find the nearest neighbor to person_features
            # distances, indices = neighbors.kneighbors(person_features.reshape(1, -1))
            distances, indices = neighbors.kneighbors(person_features)

            # Get the distance and index of the nearest neighbor
            min_distance = distances[0][0]
            min_index = indices[0][0]

            # print(f"Min feature distance = {min_distance}; {min_index}")
            print(f"Distances: {distances}")

            if min_distance <= self.feature_threshold:     # found a match  
                matching_agent_id = agent_ids[min_index]
                return matching_agent_id
            else:
                return None
        else:
            return None

    def match_features_distance(self, num_features, person_features, agent_list_features):

        feature_distances = self.feature_distances(num_features, person_features, agent_list_features)

        print(f"Feature Distances: {feature_distances}")

        if len(feature_distances) > 0: 
            min_distance = min(feature_distances, key=lambda p:p[1])
            #rospy.loginfo("min feature distance = {}".format(min_distance))
            if min_distance[1] <= self.feature_threshold:     # found a match  
                _person = next((sub for sub in self.known_agents if sub.id == min_distance[0]), None) 
                return _person.id
            else:
                return None
        else:
            return None

    '''
    Function to calculate the distances between the current histogram/texture and all know histograms/textures

    Parameters: 
    known_histograms - list of known_histograms
    new_histogram - the histogram from the current image

    Return:
    a list of agent id, distances to each know histograam
    '''
    def feature_distances(self, num_features, person_features, agent_features):
        feature_distance = []
        if person_features is not None and agent_features is not None:
            for a_feat in agent_features:
                # calculates the Euclidean distance if the number of features are the same
                if len(person_features) == len(a_feat[1]):         
                    d = euclidean_distances(person_features, a_feat[1])[0][0]
                    print(f"Feature Distance: {d}")
                    feature_distance.append((a_feat[0], d))
            
        return feature_distance    
    
    def feature_distances_normalized(self, num_features, person_features, agent_features):
        feature_distance = []
        if person_features is not None and agent_features is not None:
            for a_feat in agent_features:
                # calculates the Euclidean distance if the number of features are the same
                if len(person_features) == len(a_feat[1]):         
                    d = euclidean_distances(person_features, a_feat[1])[0][0]
                    print(f"Feature Distance: {d}")
                    feature_distance.append((a_feat[0], d))
            if len(feature_distance) > 0:
                min_distance = min(feature_distance, key=lambda p: p[1])[1]
                max_distance = max(feature_distance, key=lambda p: p[1])[1]
                
                print(f"min_distance: {min_distance}, max_distance: {max_distance}")

                # Check if max_distance - min_distance is zero
                if max_distance - min_distance == 0:
                    # Handle the case where all distances are the same
                    for i in range(len(feature_distance)):
                        feature_distance[i] = (feature_distance[i][0], 0)
                else:
                    scaler = MinMaxScaler()
                    distances = np.array(feature_distance)[:, 1].reshape(-1, 1)
                    normalized_distances = scaler.fit_transform(distances)
                    for i in range(len(feature_distance)):
                        # Normalization calculation
                        feature_distance[i] = (feature_distance[i][0], (normalized_distances[i][0] - min_distance) / (max_distance - min_distance))   
                                          
        return feature_distance

    
    def feature_distances_test(self, num_features, person_features, agent_features):
        feature_distance = []
        if person_features is not None and agent_features is not None:
            for a_feat in agent_features:
                # calculates the euclidean distance if the number of features are the same
                if len(person_features) == len(a_feat[1]):         
                    d = euclidean_distances(person_features, a_feat[1])[0][0]
                    feature_distance.append((a_feat[0], d))

            if len(feature_distance) > 0:
                min_distance = min(feature_distance, key=lambda p:p[1])[1]
                max_distance = max(feature_distance, key=lambda p:p[1])[1]           
                scaler = MinMaxScaler()
                dist = scaler.fit_transform(np.array(feature_distance)[:,1].reshape(-1,1))
                distances = np.array(dist)[:, 1].reshape(-1, 1)
                normalized_distances = scaler.fit_transform(distances)
                for i in range(len(feature_distance)):
                    feature_distance[i] = (feature_distance[i][0], (normalized_distances[i][1]-min_distance/(max_distance-min_distance)))   
                                          
        return feature_distance
    
    '''
    Determine the perspective of the person from the robot perspective 
    The perspective can only be determined if there are enough skeleton keypoints available to analyse
    
    Parameters:
    kp = skeletal keypoints. The keypoints represent joints in the human body. 
        Yolov7 Human Pose Estimation produces up to 17 skeletal keypoints

    The keypoints being used are:
    kp[0] = nose
    kp[3] = left ear
    kp[4] = right ear
    kp[5] = left shoulder
    kp[6] = right shoulder

    when the keypoint x value == 0 then it is not visible

    There are always 2 option when determining the perspective:
    1. if the shoulders and head are visible
    2. if only the head is visible

    Return:
    a string description of the perspective    
    '''
    def get_perspective(self, kp):
        # only for head
        # if (kp[4].x == 0 and kp[0].x > 0 and kp[1].x > 0 and kp[3].x > 0 and kp[3].x > kp[0].x):
        #     return "front-left" 
        # elif (kp[3].x == 0 and kp[0].x > 0 and kp[4].x > 0 and kp[0].x > kp[4].x):
        #     return "front-right"  
        # elif (kp[4].x == 0 and kp[1].x > 0 and kp[3].x > 0 and kp[3].x > kp[1].x):
        #     return "back-left"                   
        # elif (kp[3].x == 0 and kp[2].x > 0 and kp[4].x > 0 and kp[2].x > kp[4].x):
        #     return "back-right"        
        # elif (kp[3].x > 0 and kp[4].x > 0 and kp[3].x >= kp[4].x):
        #     return "front"     
        # elif (kp[3].x > 0 and kp[4].x > 0 and kp[4].x >= kp[3].x):
        #     return "back"   
        # else:
        #     return None  

        # if (kp[4].x == 0 and kp[2].x > 0 and kp[3].x > 0 and kp[3].x > kp[2].x):
        #     return "front-left" 
        # elif (kp[3].x == 0 and kp[1].x > 0 and kp[4].x > 0 and kp[1].x > kp[4].x):
        #     return "front-right"  
        # elif (kp[4].x == 0 and kp[1].x > 0 and kp[3].x > 0 and kp[3].x > kp[1].x):
        #     return "back-left"                   
        # elif (kp[3].x == 0 and kp[2].x > 0 and kp[4].x > 0 and kp[2].x > kp[4].x):
        #     return "back-right"        
        # elif (kp[3].x > 0 and kp[4].x > 0 and kp[3].x >= kp[4].x):
        #     return "front"     
        # elif (kp[3].x > 0 and kp[4].x > 0 and kp[4].x >= kp[3].x):
        #     return "back"   
        # else:
        #     return None          
    
        if (kp[4].x == 0 and kp[2].x > 0 and kp[3].x > 0 and kp[3].x > kp[2].x):
            return "front-left" 
        elif (kp[3].x == 0 and kp[1].x > 0 and kp[4].x > 0 and kp[1].x > kp[4].x):
            return "front-right"       
        elif (kp[3].x > 0 and kp[4].x > 0 and kp[3].x >= kp[4].x):
            return "front"     
        elif (kp[3].x > 0 and kp[4].x > 0 and kp[4].x >= kp[3].x):
            return "back"  
        elif (kp[4].x == 0 and kp[3].x > 0):
            return "back-left"                   
        elif (kp[3].x == 0 and kp[4].x > 0):
            return "back-right"           
        else:
            return None  

    def get_perspective_old(self, kp):
        if (kp[4].x == 0 and kp[0].x > 0 and kp[3].x > 0 and kp[3].x > kp[0].x) or (kp[4].x == 0 and kp[5].x > 0 and kp[6].x > 0 and kp[5].x > kp[6].x):
            return "front-left" 
        elif (kp[3].x == 0 and kp[5].x > 0 and kp[6].x > 0 and kp[5].x > kp[6].x) or (kp[3].x == 0 and kp[0].x > 0 and kp[4].x > 0 and kp[0].x > kp[4].x):
            return "front-right"  
        elif (kp[4].x == 0 and kp[6].x == 0 and kp[5].x > 0 and kp[1].x > 0 and kp[5].x > kp[1].x) or (kp[4].x == 0 and kp[1].x > 0 and kp[3].x > 0 and kp[3].x > kp[1].x):
            return "back-left"                   
        elif (kp[3].x == 0 and kp[5].x == 0 and kp[4].x > 0 and kp[6].x > 0 and kp[6].x > kp[4].x) or (kp[3].x == 0 and kp[2].x > 0 and kp[4].x > 0 and kp[2].x > kp[4].x):
            return "back-right"        
        elif (kp[5].x > 0 and kp[6].x > 0 and kp[5].x > kp[6].x) or (kp[3].x > 0 and kp[4].x > 0 and kp[3].x > kp[4].x):
            return "front"     
        elif (kp[5].x > 0 and kp[6].x > 0 and kp[6].x > kp[5].x) or (kp[3].x > 0 and kp[4].x > 0 and kp[4].x > kp[3].x):
            return "back"   
        else:
            return None   

    '''
    Add the person to the list of known agents.
    Also add the agent to the world model if the flag has been set to true

    Parameters:
    person  - the new agent that we need to add. the data will represent face/features for the observed perspective
    '''
    def add_agent(self, person):
        print("Adding new agent")

        if person.id is None:
            person.id  = len(self.known_agents)+1
        
        self.known_agents.append(person)

        # self.print_known_agents()
        if self.use_world_model:
            self.agent.insert_agent(self.db.con_pool, "P", person)
         
    '''
    Update an existing agent in the list of known agents.
    Also update the agent to the world model if the flag has been set to true

    Parameters:
    person  - the spefic agent that we need to update. the data will represent face/features for the observed perspective
    '''
    def update_agent(self, person):
        for knwn_agent in self.known_agents:
            if knwn_agent.id == person.id:
                print(f"Updating Agent {knwn_agent.id}")
                self.agent.update_agent_last_seen(self.db.con_pool, person.id, person.position)
                # check if feature currently exists        
                for new_feat in person.features:
                    found_features = False
                    if new_feat.encoding is not None:
                        for known_feat in knwn_agent.features:                                          
                            if known_feat.perspective == new_feat.perspective and known_feat.feature_type == new_feat.feature_type:
                                # rospy.loginfo(f"Found existing feature for {person.id}: {new_feat.perspective} {new_feat.feature_type}")
                                known_feat.encoding = new_feat.encoding
                                found_features = True
                    
                        # if it is a new perspective/feature add it                   
                        if found_features == False:
                            # rospy.loginfo(f"Adding new feature for {person.id}: {new_feat.perspective} {new_feat.feature_type}")
                            feat = Feature(new_feat.perspective, new_feat.feature_type, new_feat.encoding)
                            knwn_agent.add_feature(feat)
                       
                            if self.use_world_model:
                                self.agent.insert_agent_feature(self.db.con_pool, person.id, new_feat)
                        else:
                            if self.use_world_model:
                                self.agent.update_agent_feature(self.db.con_pool, person.id, new_feat)
                         
    '''
    Calculate the euclidean distance between 2 keypoints
    Keypoints contain X, Y values
    
    Parameters:
    kp1 = the first keypoint 
    kp2 = the second keypoint

    Return:
    d = distance between
    '''
    def kp_dist(self, kp1, kp2):
        d = sqrt(pow((kp1.x - kp2.x),2) + pow((kp1.y - kp2.y),2))
        return d

    '''
    This function takes a small crop of the shirt area around the specified perspective.
    
    Parameters:
    img = the image frame being examined
    kp = skeletal keypoints. The keypoints represent joints in the human body. 
        Yolov7 Human Pose Estimation produces up to 17 skeletal keypoints
    perspective = the perspective that has been detected

    Note:
    the keypoint is visible when the kp.x > 0 
    if kp.x = 0 then we will not use it as it is too close to the border

    Returns:
    if the region of interest has been detected (i.e. keypont is visible)
        the cropped region of interest
    else
        None
    '''
    def crop_shirt(self, img, kp, perspective): 
        pixel_from_border = 2   
        crop_width = self.crop_size  
        img_height, img_width = img.shape[:2]
        x, y, w, h = None, None, None, None
       # self.print_keypoints(kp)
        if perspective == "front":
            if kp[6].x > 0 and kp[11].x > 0:
                x = int(kp[6].x)
                y = int(kp[6].y)
                w = int(kp[11].x - x)
                h = int(kp[11].y - y)
            elif kp[5].x > 0 and kp[6].x > 0:
                x = int(kp[6].x)
                y = int(kp[6].y)
                w = int(kp[5].x - x)
                h = int(img_height - y - pixel_from_border) 
            elif kp[11].x > 0 and kp[12].x > 0:
                x = int(kp[12].x)
                y = int(kp[12].y)
                w = int(kp[11].x - x)
                h = int(y)                 
        elif perspective == "front-left":
            if kp[5].x > 0 and kp[7].x > 0:
                if int(kp[5].x-crop_width) > 0:
                    x = int(kp[5].x-crop_width)
                else:
                    x = 0
                y = int(kp[5].y)
                w = int(kp[7].x+5 - x)
                h = int(kp[7].y - y)
            elif kp[5].x > 0:
                if int(kp[5].x-crop_width) > 0:
                    x = int(kp[5].x-crop_width)
                else:
                    x = 0                
                y = int(kp[5].y)
                w = int(kp[5].x+5 - x)
                h = int(img_height - 2 - y) 
            elif kp[7].x > 0:
                if int(kp[7].x-crop_width) > 0:
                    x = int(kp[7].x-crop_width)
                else:
                    x = 0                
                y = int(kp[7].y)
                w = int(kp[7].x+5 - x)
                h = int(y)  
        elif perspective == "back":
            if kp[5].x > 0 and kp[12].x > 0:
                x = int(kp[5].x)
                y = int(kp[5].y)
                w = int(kp[12].x - x)
                h = int(kp[12].y - y)
            elif kp[5].x > 0 and kp[6].x > 0:
                x = int(kp[5].x)
                y = int(kp[5].y)
                w = int(kp[6].x - x)
                h = int(img_height - y - pixel_from_border) 
            elif kp[11].x > 0 and kp[12].x > 0:
                x = int(kp[11].x)
                y = int(kp[11].y)
                w = int(kp[12].x - x)
                h = int(y)   
        elif perspective == "front-right":
            if kp[6].x > 0 and kp[8].x > 0:
                if int(kp[6].x-crop_width) > 0:
                    x = int(kp[6].x-crop_width)
                else:
                    x = 0
                y = int(kp[6].y)
                w = int(kp[8].x+5 - x)
                h = int(kp[8].y - y)
            elif kp[6].x > 0:
                if int(kp[6].x-crop_width) > 0:
                    x = int(kp[6].x-crop_width)
                else:
                    x = 0                
                y = int(kp[6].y)
                w = int(kp[6].x+5 - x)
                h = int(img_height - 2 - y) 
            elif kp[8].x > 0:
                if int(kp[8].x-crop_width) > 0:
                    x = int(kp[8].x-crop_width)
                else:
                    x = 0                
                y = int(kp[8].y)
                w = int(kp[8].x+5 - x)
                h = int(y)          
        elif perspective == "back-left":
            if kp[5].x > 0 and kp[11].x > 0:
                x = int(kp[5].x)
                y = int(kp[5].y)
                if int(kp[5].x + crop_width) <= img_width:
                    w = int(kp[5].x + crop_width)
                else:
                    w = int(img_width - kp[5].x) 
                h = int(kp[11].y - kp[5].y)
            elif kp[5].x > 0 and kp[11].x == 0:
                x = int(kp[5].x)
                y = int(kp[5].y)
                if int(kp[5].x + crop_width) <= img_width:
                    w = int(kp[5].x + crop_width)
                else:
                    w = int(img_width - kp[5].x) 
                h = int(img_height - y - pixel_from_border)   
            elif kp[5].x == 0 and kp[11].x > 0:
                x = int(kp[11].x)
                y = int(kp[11].y)
                w = int(kp[11].x + crop_width)
                h = int(y) 
        elif perspective == "back-right":
            if kp[6].x > 0 and kp[12].x > 0:
                x = int(kp[6].x)
                y = int(kp[6].y)
                if int(kp[6].x + 20) <= img_width:
                    w = int(kp[6].x + crop_width)
                else:
                    w = int(img_width - kp[6].x)
                h = int(kp[12].y - kp[6].y)
            elif kp[6].x > 0 and kp[12].x == 0:
                x = int(kp[6].x)
                y = int(kp[6].y)
                if int(kp[6].x + crop_width) <= img_width:
                    w = int(kp[6].x + crop_width)
                else:
                    w = int(img_width - kp[6].x)
                h = int(img_height - y - pixel_from_border)   
            elif kp[6].x == 0 and kp[12].x > 0:
                x = int(kp[12].x)
                y = int(kp[12].y)
                if int(kp[12].x + crop_width) <= img_width:
                    w = int(kp[12].x + crop_width)
                else:
                    w = int(img_width - kp[12].x) 
                h = int(y)                              
        else:   # did not have perspective
            return None
        
        if x is None or y is None or w is None or h is None:
            return None
        
        if w < 20 or h < 20:
            return None
        
        return img[y:y+h,x:x+w]

    '''
    This function takes a small crop of the head area around the specified perspective.
    
    Parameters:
    img = the image frame being examined
    bbox_head = the bounding box of the head
    perspective = the perspective that has been detected

    Returns:
    if the region of interest has been detected (i.e. head is visible)
        the cropped region of interest
    else
        None
    '''
    def crop_roi(self, img, bbox_head, perspective):   
        try:
            img_height, img_width = img.shape[:2]
            # x, y, w, h = None, None, self.crop_w, self.crop_h
            offset = 3
            h_multiplier = self.h_multiplier 
            # want to make the crop width a few pixels smaller on each side to minimise potential noise
            # define the height of the crop
            # if the bounding box for the head is available then use it to find the top of the head
            if bbox_head is not None and perspective is not None:

                # define top and height of crop
                y = bbox_head.y
                h = int((bbox_head.width - (2 * offset)) * h_multiplier)
                # define the width of the crop
                x = bbox_head.x + offset 
                w = bbox_head.width - (2 * offset)  
            else:
                return None
            
            if x is None or y is None:
                return None      

            # save the image for debugging
            if img[y:y+h,x:x+w].size > 0:
                if self.save_img:            
                    # save the cropped image to a file
                    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S_%f")[:-3]
                    filename = f"crop_{perspective}_{timestamp}.jpg"
                    # Define the directory
                    directory = Path(self.img_path)

                    # Ensure the directory exists
                    directory.mkdir(parents=True, exist_ok=True)

                    # Write the image            
                    cv2.imwrite(str(directory / filename), img[y:y+h,x:x+w])   
                    print(f"Saved image: {x}, {y}, {w}, {h} - {filename}") 

                return img[y:y+h,x:x+w] 
            else:
                return None
        except Exception as e:
            rospy.logwarn(f"Crop head Error:{perspective}, {img_height}, {img_width}, {x}, {y}, {w}, {h}, {e}, {kp} ")
            return None
        

    def crop_roi_old(self, img, bbox_person, bbox_head, kp, perspective):   
        try:
            img_height, img_width = img.shape[:2]
            x, y, w, h = None, None, self.crop_w, self.crop_h
            crop_w = self.crop_w      # number of pixels to crop
            offset = 12
            w_offset = 3
            h_multiplier = self.h_multiplier 
            bbox_type = None
            # want to make the crop width a couple of pixels smaller on each side to minimise potential noise
            # crop width = (bbox_head.x + w_offset) - (bbox_head.width - (2 * w_offset))
            # define the height of the crop
            # if the bounding box for the head is available then use it to find the top of the head
            # else use the bounding box of the person
            if bbox_head is not None:
                y = bbox_head.y
                h = int((bbox_head.width - (2 * w_offset)) * h_multiplier)
                # h = int(bbox_head.width * h_multiplier)
            else:
                y = bbox_person.y
                h = self.crop_h

            if y < 0:
                y = 0
                if y + int(kp[12].y) >= img_height:
                    h = y + self.crop_h - img_height                

            # define the width of the crop
            if perspective == "front":
                if bbox_head.width > 0:     # can see the head
                    # x = bbox_head.x + w_offset
                    # w = bbox_head.width 
                    x = bbox_head.x + w_offset 
                    w = bbox_head.width - (2 * w_offset)  
                    bbox_type = "head"
                elif kp[3].x > 0 and kp[4].x > 0:     # can see left ear and right ear
                    x = int(kp[4].x) - int(offset)        # we will take crop from the right ear  

                    if x < 0:
                        x = 0
                        w = crop_w
                    elif x + crop_w > img_width:
                        w = x + crop_w - img_width 
                    else:
                        w = int(kp[3].x) - x     
                    bbox_type = "kp"          
            elif perspective == "front-left":
                if bbox_head.width > 0:     # can see the head
                    # x = bbox_head.x + w_offset
                    # w = bbox_head.width
                    x = bbox_head.x + w_offset 
                    w = bbox_head.width - (2 * w_offset)   
                    bbox_type = "head"                  
                elif kp[2].x > 0 and kp[3].x > 0:     # can see right eye and left ear
                    x = int(kp[2].x) - int(offset*2)
                    if x < 0:
                        x = 0
                        w = crop_w
                    elif x + crop_w > img_width:
                        w = x + crop_w - img_width 
                    else:
                        w = int(kp[3].x) + int(offset/2) - x   
                    bbox_type = "kp"                                   
            elif perspective == "back":
                if bbox_head.width > 0:     # can see the head
                    # x = bbox_head.x + w_offset
                    # w = bbox_head.width
                    x = bbox_head.x + w_offset 
                    w = bbox_head.width - (2 * w_offset)     
                    bbox_type = "head"                            
                elif kp[3].x > 0 and kp[4].x > 0:     # can see left ear and right ear
                    x = int(kp[3].x)              
                    if x < 0:
                        x = 0
                        w = crop_w
                    elif x + crop_w > img_width:
                        w = x + crop_w - img_width 
                    else:
                        w = int(kp[4].x) - x    
                    bbox_type = "kp"             
            elif perspective == "front-right":
                if bbox_head.width > 0:     # can see the head
                    # x = bbox_head.x + w_offset
                    # w = bbox_head.width
                    x = bbox_head.x + w_offset 
                    w = bbox_head.width - (2 * w_offset)     
                    bbox_type = "head"          
                elif kp[4].x > 0 and kp[1].x > 0:    # can see right ear and left eye
                    x = int(kp[4].x) - int(offset) 
                    if x < 0:
                        x = 0
                        w = crop_w
                    elif x + crop_w > img_width:
                        w = x + crop_w - img_width 
                    else:
                        w = int(kp[1].x) + int(offset) - x    
                    bbox_type = "kp"               
            elif perspective == "back-left":
                if bbox_head.width > 0:     # can see the head
                    # x = bbox_head.x + w_offset
                    # w = bbox_head.width
                    x = bbox_head.x + w_offset 
                    w = bbox_head.width - (2 * w_offset)
                    bbox_type = "head"
                elif kp[3].x > 0:     # can see left ear
                    x = int(kp[3].x) - int(offset)
                    if x < 0:
                        x = 0
                        w = crop_w
                    elif x + crop_w > img_width:
                        w = x + crop_w - img_width 
                    else:
                        w = int(kp[3].x) + int(offset*2) - x 
                    bbox_type = "kp"
            elif perspective == "back-right":
                if bbox_head.width > 0:     # can see the head
                    # x = bbox_head.x + w_offset
                    # w = bbox_head.width
                    x = bbox_head.x + w_offset 
                    w = bbox_head.width - (2 * w_offset) 
                    bbox_type = "head"          
                elif kp[4].x > 0:     # can see right ear 
                    x = int(kp[4].x) - int(offset*2)     
                    if x < 0:
                        x = 0
                        w = crop_w
                    elif x + crop_w > img_width:
                        w = x + crop_w - img_width 
                    else:
                        w = int(kp[4].x) + int(offset) - x
                    bbox_type = "kp"
            else:   # did not have perspective
                return None
            
            if x is None or y is None:
                return None      

            if img[y:y+h,x:x+w].size > 0:
                # norm_img = self.histogram_equalization(img[y:y+h,x:x+w])
                # grayscale_img = cv2.cvtColor(img[y:y+h,x:x+w], cv2.COLOR_BGR2GRAY)
                # save the image for debugging
                if self.save_img:            
                    # save the cropped image to a file
                    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S_%f")[:-3]
                    filename = f"crop_{perspective}_{timestamp}.jpg"
                    # Define the directory
                    directory = Path.home() / "workspace/vision_ws/src/vision_pkgs/person_recognition/images"

                    # Ensure the directory exists
                    directory.mkdir(parents=True, exist_ok=True)

                    # Write the image            
                    cv2.imwrite(str(directory / filename), img[y:y+h,x:x+w])   
                    # cv2.imwrite(str(directory / f"gray_{filename}"), grayscale_img)
                    print(f"Saved image: {bbox_type}, {x}, {y}, {w}, {h} - {filename}") 
                    # print(f"Saved image: norm_{filename}")

                # return norm_img 
                return img[y:y+h,x:x+w] 
                # return grayscale_img
            else:
                return None
        except Exception as e:
            rospy.logwarn(f"Crop head Error:{perspective}, {img_height}, {img_width}, {x}, {y}, {w}, {h}, {e}, {kp} ")
            return None

    def crop_head_old(self, img, bbox_person, bbox_head, kp, perspective):   
        try:
            img_height, img_width = img.shape[:2]
            x, y, w, h = None, None, self.crop_w, self.crop_h
            crop_w = self.crop_w      # number of pixels to crop
            offset = 12

            # define the height of the crop
            # if the bounding box for the head is available then use it to find the top of the head
            # else use the bounding box of the person
            if bbox_head.y is not None:
                y = bbox_head.y 
            else: 
                y = bbox_person.y

            if y < 0:
                y = 0
                h = self.crop_h
            if y + int(kp[12].y) >= img_height:
                h = y + self.crop_h - img_height                
            
            # define the width of the crop
            if perspective == "front":
                if bbox_head.width > 0:     # can see the head
                    x = bbox_head.x
                    w = bbox_head.width    
                elif kp[3].x > 0 and kp[4].x > 0:     # can see left ear and right ear
                    x = int(kp[4].x) - int(offset)        # we will take crop from the right ear  

                    if x < 0:
                        x = 0
                        w = crop_w
                    elif x + crop_w > img_width:
                        w = x + crop_w - img_width 
                    else:
                        w = int(kp[3].x) - x               
            elif perspective == "front-left":
                if bbox_head.width > 0:     # can see the head
                    x = bbox_head.x
                    w = bbox_head.width                   
                elif kp[2].x > 0 and kp[3].x > 0:     # can see right eye and left ear
                    x = int(kp[2].x) - int(offset*2)
                    if x < 0:
                        x = 0
                        w = crop_w
                    elif x + crop_w > img_width:
                        w = x + crop_w - img_width 
                    else:
                        w = int(kp[3].x) + int(offset/2) - x                                      
            elif perspective == "back":
                if bbox_head.width > 0:     # can see the head
                    x = bbox_head.x
                    w = bbox_head.width                                    
                elif kp[3].x > 0 and kp[4].x > 0:     # can see left ear and right ear
                    x = int(kp[3].x)              
                    if x < 0:
                        x = 0
                        w = crop_w
                    elif x + crop_w > img_width:
                        w = x + crop_w - img_width 
                    else:
                        w = int(kp[4].x) - x                 
            elif perspective == "front-right":
                if bbox_head.width > 0:     # can see the head
                    x = bbox_head.x
                    w = bbox_head.width                
                elif kp[4].x > 0 and kp[1].x > 0:    # can see right ear and left eye
                    x = int(kp[4].x) - int(offset) 
                    if x < 0:
                        x = 0
                        w = crop_w
                    elif x + crop_w > img_width:
                        w = x + crop_w - img_width 
                    else:
                        w = int(kp[1].x) + int(offset) - x                   
            elif perspective == "back-left":
                if bbox_head.width > 0:     # can see the head
                    x = bbox_head.x
                    w = bbox_head.width
                elif kp[3].x > 0:     # can see left ear
                    x = int(kp[3].x) - int(offset)
                    if x < 0:
                        x = 0
                        w = crop_w
                    elif x + crop_w > img_width:
                        w = x + crop_w - img_width 
                    else:
                        w = int(kp[3].x) + int(offset*2) - x 
            elif perspective == "back-right":
                if bbox_head.width > 0:     # can see the head
                    x = bbox_head.x
                    w = bbox_head.width                
                elif kp[4].x > 0:     # can see right ear 
                    x = int(kp[4].x) - int(offset*2)     
                    if x < 0:
                        x = 0
                        w = crop_w
                    elif x + crop_w > img_width:
                        w = x + crop_w - img_width 
                    else:
                        w = int(kp[4].x) + int(offset) - x
            else:   # did not have perspective
                return None
            
            if x is None or y is None:
                return None      

            # save the image for debugging
            if self.save_img:            
                # save the cropped image to a file
                timestamp = datetime.now().strftime("%Y%m%d_%H%M%S_%f")[:-3]
                filename = f"crop_{perspective}_{timestamp}.jpg"
                # Define the directory
                directory = Path.home() / "workspace/vision_ws/src/vision_pkgs/person_recognition/images"

                # Ensure the directory exists
                directory.mkdir(parents=True, exist_ok=True)

                # Write the image
                cv2.imwrite(str(directory / filename), img[y:y+h,x:x+w])   
                print(f"Saved image: {filename}") 

            return img[y:y+h,x:x+w] 
        except Exception as e:
            rospy.logwarn(f"Crop head Error:{perspective}, {img_height}, {img_width}, {x}, {y}, {w}, {h}, {e}, {kp} ")
            return None

    # CNN
    def get_feature(self, img):
        # resize the image in case it has been adjusted from the crop
        resized_img = cv2.resize(img, (self.crop_w, self.crop_h))
        # Load and preprocess the image
        resized_img = preprocess_input(resized_img)
        features = self.feature_extractor.predict(np.expand_dims(resized_img, axis=0))
        # normalize the feature vector
        normalized_features = features / np.linalg.norm(features)
        return normalized_features
    
    # LBP
    def get_feature_lbp(self, img, points=8, radius=1):
        """Calculate Local Binary Pattern (LBP) of the image"""
        hist = None
        if img is not None:
            if img.size > 0:
                gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
                lbp = feature.local_binary_pattern(gray, points, radius, method="uniform")
                hist, _ = np.histogram(lbp.ravel(), bins=np.arange(0, points + 3), range=(0, points + 2))
                hist = hist.astype("float")
                hist /= (hist.sum() + 1e-8)  # Normalize the histogram, 1e-8 to avoid divide by zero
        
        return hist

    def feature_distances_lbp(self, agent_features,  person_features):
        feature_distance = []
        for a_feat in agent_features:
            for p_feat in person_features:
                feature_distance.append((a_feat[0], np.linalg.norm(p_feat - a_feat[1])))
                
        return feature_distance
      
    # SIFT
    def get_feature_sift(self, img):
        """Calculate SIFT of the image"""
        if img is not None:            
            if img.size > 0:
                gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
                sift = cv2.SIFT_create()
                # Detect keypoints and compute descriptors for both images
                _, descriptors = sift.detectAndCompute(gray, None)

        return descriptors    


    def get_feature_sift_old(self, img, points=8, radius=1):
        """Calculate SIFT of the image"""
        descriptors = "test"
        if img is not None:            
            if img.size > 0:
                gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
                sift = cv2.SIFT_create()
                # Detect keypoints and compute descriptors for both images
                _keypoints, descriptors = sift.detectAndCompute(gray, None)
        print(f"Descriptors: {descriptors}")
        return descriptors.ravel()    

    def histogram_equalization(self, img):
        b, g, r = cv2.split(img)
        b_equalized = cv2.equalizeHist(b)
        g_equalized = cv2.equalizeHist(g)
        r_equalized = cv2.equalizeHist(r)

        return cv2.merge((b_equalized, g_equalized, r_equalized))

    def sharpen_image(self, img):
        kernel = np.array([[-1,-1,-1], [-1,9,-1], [-1,-1,-1]])

        sharpened_image = np.clip(sharpened_image, 0, 255)
        return cv2.filter2D(img, -1, kernel)

    '''
    An image is generally scaled prior to processing to improve performance.
    This function rescales the bounding box coordinates to match the size of the original image

    Parameters:
    x = the x coordinate of the scaled bounding box
    y = the y coordinate of the scaled bounding box
    width = the width of the scaled bounding box
    height = the height of the scaled bounding box
    scale = the scaling factor

    Return:
    rescaled bounding box
    '''
    def rescale_bbox(self, x, y, width, height, scale):
        bbox = BoundingBox()
        bbox.width = int(width*(1/scale))
        bbox.height = int(height*(1/scale))
        bbox.x = int(x*(1/scale))
        bbox.y = int(y*(1/scale))

        return bbox

    '''
    Function to calculate the distances between the new face encoding and all known face encodings
    This gets the distances for face encodings for each perspective so there may be multiple rows for each agent 
    depending on how many perspectives of the face we have

    Parameters: 
    new_face - the new face from the current image

    Return:
    a list of distance to each known face
    a list of index to known_agentto each known face
    '''
    def face_distance(self, new_face):
        face_distances = [] 
        agent_list = []
        result = None
        
        if new_face is not None:                                  
            for idx, agent in enumerate(self.known_agents):
                
                for agt_feat in agent.features:
                    if agt_feat.feature_type == "face" and agt_feat.encoding is not None:
                        # rospy.loginfo(f"{agent.id} memory face {agt_feat.perspective}")                
                        #face_distances.append((agent.id, np.linalg.norm(agt_feat.encoding - new_face)))
                        face_distances.append(agt_feat.encoding)
                        agent_list.append(idx)
                        
            result = face_recognition.api.face_distance(face_distances, new_face)
            
        return result, agent_list                       

    '''
    Displays the image in a window 
    '''
    def display_person(self, img, bbox):
        border_size=10

        #rospy.loginfo(f"Img = {len(img)}; 1 = {bbox.y-border_size}; 2 = {bbox.y+((bbox.height+border_size)-(bbox.y-border_size))}; 3 = {(bbox.x-border_size)}; 4 = {bbox.y+((bbox.width+border_size)-(bbox.x-border_size))}")
        #rospy.loginfo(f"Img = {len(img)}; BBox.X={bbox.x}, BBoxY={bbox.y}; BBoxW={bbox.width}; BBoxH={bbox.height}; Y = {(bbox.x+border_size)}; H = {(bbox.width-border_size)}; X = {(bbox.y+border_size)}; W = {(bbox.height-border_size)}")
        if bbox is not None:
            #cropped_img = img[(bbox.y-border_size):bbox.y+((bbox.height+border_size)-(bbox.y-border_size)), (bbox.x-border_size):bbox.y+((bbox.width+border_size)-(bbox.x-border_size))] 
            #image[x:w, y:h]
            #cropped_img = img[:(bbox.width-(2*border_size)), (bbox.y+border_size): (bbox.height-(2*border_size))] 
            # start_y: end_y, start_x: end_x
            #cropped_img = img[(bbox.y+border_size):(bbox.height-(border_size)), (bbox.x+border_size):(bbox.width-(border_size))] 
            cropped_img = img[(bbox.y+border_size):(bbox.height-(border_size)), (bbox.x+border_size):(bbox.width-(border_size))]
            rospy.loginfo(f"Cropped img = {len(cropped_img)}")
            if np.any(cropped_img):
                cv2.imshow('Display Faces', cropped_img)

        # Hit 'q' on the keyboard to quit!
        if cv2.waitKey(1) & 0xFF == ord('q'):
            cv2.destroyAllWindows()
    
    def print_known_agents(self):

        for agent in self.known_agents:
            agent.print_features()
            # agent.print_feature_details()
