#!/usr/bin/env python3

# Person Recoginition 
# Detects people using texture and faces and captures the name and stores in the world model database
# Libraries:
# face recognition - https://github.com/ageitgey/face_recognition

# Dependencies:
# World Model - database with schema 2.1.4 or higher which contains the agent table structures - use with flag: use_world_model = True

# Written By: Adam Golding

#from deepface import DeepFace
from datetime import datetime
from pathlib import Path
import rospy
import sys
import os
from math import sqrt
import torch    
from sensor_msgs.msg import Image
from unsw_vision_msgs.msg import DetectionList, PersonDetection, BoundingBox
import cv2
from cv_bridge import CvBridge
import numpy as np
# from skimage import feature
# from scipy.spatial import distance
# from scipy.stats import chisquare
# from visualizer import draw_face_bbox, draw_face_name, draw_face_info_from_msg, draw_detections, plot_skeleton_kpts, print_objects, display_image, visualize_image, print_agent, print_known_agents
from visualizer import display_image, visualize_image, print_agent, print_known_agents
module_path = os.environ.get("UNSW_WS")
sys.path.append(module_path + "/WORLD_MODEL/world_model/world_model_kb/scripts")
# from agent_api import AgentApi
# from dbase import DBase
from unsw_person_recognition import PersonRecognition
import face_recognition


class PersonRecogniser():
    def __init__(self) -> None:
        # parameters
        self.input_topic = rospy.get_param('~input_topic')                  # topic to get the image from
        self.output_topic = rospy.get_param('~output_topic')                # topic to publish the face detection info to            
        self.vis_output_topic = rospy.get_param('~vis_output_topic')        # topic to publish the visualization of the detections         
        self.save_img = rospy.get_param('~save_img', False)                 # flag indicates whether to save the image
        self.img_path = rospy.get_param('~img_path')                        # path where face image is saved
        self.display_img = rospy.get_param('~display_img',False)            # flag to indicate whether to display the image with bounding box and person name
        self.use_world_model = rospy.get_param('~use_world_model',False)    # flag to indicate whether to use the world model database
        self.db_ini = rospy.get_param('~db_ini')                            # file containing database connection details
        self.border_size = rospy.get_param('~border_size')                  # the size of the border for the bounding box cropped image
        self.use_gpu = rospy.get_param('~use_gpu', False)                   # flag indicating whether to use a gpu for processing the image
        self.visualize = rospy.get_param('~visualize', False)               # flag indicating whether to publish a visualization of the detections
        self.queue_size = rospy.get_param('~queue_size')                    # the queue size for the publishers
        self.scale = float(rospy.get_param('~scale'))                       # the amount to scale the image, must be a float, e.g. 0.25, 1.0
        self.face_threshold = float(rospy.get_param('~face_threshold'))   # the threshold to determine if comparing 2 faces are the same
        self.feature_threshold = float(rospy.get_param('~feature_threshold'))   # the threshold to determine if comparing 2 histograms are the same
        # self.deepface_model = rospy.get_param('~deepface_model')             # the face recognition model to use
        self.frame_count = 0                                                # initialise the frame count so we know how many frames we have processed

        rospy.loginfo("Initialising Person Recognition")

        self.pr = PersonRecognition(self.db_ini, self.scale, self.feature_threshold, self.face_threshold, self.use_world_model, self.save_img, self.img_path)
        # log info
        rospy.loginfo("input_topic="+self.input_topic)
        rospy.loginfo("output_topic="+self.output_topic)
        rospy.loginfo("vis_output_topic="+self.vis_output_topic)                
        rospy.loginfo("display_img="+str(self.display_img))
        rospy.loginfo("img_path="+str(self.img_path))
        rospy.loginfo("display_img="+str(self.display_img))
        rospy.loginfo("use_world_model="+str(self.use_world_model))
        rospy.loginfo("use_gpu="+str(self.use_gpu))
        rospy.loginfo("db_ini="+self.db_ini)
        rospy.loginfo("visualize="+str(self.visualize))
        rospy.loginfo("queue_size="+str(self.queue_size))       
        rospy.loginfo("scale="+str(self.scale))
        rospy.loginfo("face_threshold="+str(self.face_threshold))
        rospy.loginfo("feature_threshold="+str(self.feature_threshold))
        # rospy.loginfo("deepface_model="+str(self.deepface_model))
        print("\n")

        # Create an empty RGB image
        empty_image = np.zeros((100, 100, 3), dtype=np.uint8)

        # Call face_locations with the empty image
        face_recognition.face_locations(empty_image, model="cnn")


        # subscribers
        self.image_listener = rospy.Subscriber(self.input_topic, DetectionList, self.recognise) # subscribe to the img topic

        if self.visualize:
            # The publisher for the visualization of the detections
            self.vis_publisher = rospy.Publisher(self.vis_output_topic, Image, queue_size=self.queue_size) 

        # publishers
        self.face_publisher = rospy.Publisher(self.output_topic, DetectionList, queue_size=self.queue_size)     

        # other
        self.bridge = CvBridge()

        # Load the pre-trainined DeepFace model
        #self.df_model = DeepFace.build_model(self.deepface_model)



        rospy.loginfo("Person Recogniser started")   
        rospy.loginfo("\n")

    '''   
    Callback Method 
    Method processes each image frame to determine, record and publish person information in the vision message
    '''
    def recognise(self, in_msg): 
        self.frame_count += 1
        print(f"\n**New Image: {self.frame_count}")        
        ### Image Preparation
        img = self.bridge.imgmsg_to_cv2(in_msg.image, desired_encoding="bgr8")   
        if self.save_img:
            # save the image to a file
            timestamp = datetime.now().strftime("%Y%m%d_%H%M%S_%f")[:-3]
            filename = f"img_{self.frame_count}_{timestamp}.jpg"
            # Define the directory
            directory = Path(self.img_path)

            # Ensure the directory exists
            directory.mkdir(parents=True, exist_ok=True)

            # Write the image
            cv2.imwrite(str(directory / filename), img)   
            print(f"Saved image: {filename}")              
        # if we have people detected in the image then we will continue to process the image
        if len(in_msg.people) > 0:    
            # Resize frame of video to the scale size for faster face recognition processing
            # The larger the scaling (i.e. smaller fraction) the less faces will be detected and 
            # the closer the people will need to be to the camera. 
            # the scale size is set in the launch file
            small_frame = cv2.resize(img, (0, 0), fx=self.scale, fy=self.scale)
                
            # Convert the image from BGR color (which OpenCV uses) to RGB color (which face_recognition uses)
            rgb_small_frame = np.ascontiguousarray(small_frame[:, :, ::-1])

            # Find all the faces and face encodings in the current frame of video
            #face_locations = DeepFace.extract_faces(rgb_small_frame,detector_backend="opencv")

            if self.use_gpu:
                face_locations = face_recognition.face_locations(rgb_small_frame, model="cnn")
            else:
                face_locations = face_recognition.face_locations(rgb_small_frame)
            face_encodings = face_recognition.face_encodings(rgb_small_frame, face_locations, num_jitters=5)
            self.face_publisher.publish(in_msg)
            print(f"People in image: {len(in_msg.people)}; Faces detected: {len(face_encodings)}; Known People: {len(self.pr.known_agents)}")
            
            person_list = []    # list of people found in the image

            for person in in_msg.people:
                _person = self.pr.extract_person_features(img, person, face_locations, face_encodings)      # extract the features of the poi, if we did not find a perspective the None is returned        
                if _person is not None:     # i.e. a person could be extracted because we have identified the perspective                    
                    _person = self.pr.match_person(_person)  # match the person to known people/agents, returns the matched person                        
                    # print(f"Matched person: {_person}")
                    if _person.features is not None:    # need to have features of a person before we add them, this could be either face or other features
                        if _person.id is None:
                            # add agent 
                            self.pr.add_agent(_person)
                            # if _person.bbox_face is not None:
                                # self.pr.display_person(img, _person.bbox_face)
                        else:
                            # update agent 
                            self.pr.update_agent(_person)

                        # add to the list of people in the current frame
                        person_list.append(_person)

            new_msg = self.create_output_message(in_msg, person_list)
            # publish the new message to output topic
            self.face_publisher.publish(new_msg)

            # display the image to a window
            # if self.display_img:           
            #     display_image(img, person_list)

            if self.visualize:
                out_msg = visualize_image(img, new_msg)
                self.vis_publisher.publish(out_msg)  

        else:
            # republish the original message to output topic
            self.face_publisher.publish(in_msg)       


    def create_output_message(self, in_msg, person_list):
        # copy the Detection List message and update the people information
        out_msg = DetectionList()
        out_msg.header = in_msg.header
        out_msg.image = in_msg.image
        out_msg.objects = in_msg.objects

        # add updated people
        # TODO: need to fix the message creation with various components of the message
        for person in person_list:
            #print(f"Person: {person}")
            person_msg = PersonDetection()
            if person.id is not None:
                person_msg.id = str(person.id)
            if person.name is not None:                
                person_msg.name = str(person.name)
            if person.bbox_face is not None:
                person_msg.bbox_face = person.bbox_face
            if person.bbox_person is not None:
                person_msg.bbox_person = person.bbox_person
            # if person.get("face_encoding"):
            #     person_msg.face_encoding = person["face_encoding"]
            # if person.get("position"):
            #     person_msg.position = personp["position"]
            # if person.get("face_position"):
            #     person_msg.face_position = person["face_position"]
            # if person.get("centroid"):
            #     person_msg.centroid = person["centroid"]
            if person.skeleton is not None:
                person_msg.skeleton = person.skeleton
            # person_features = person.print_feature_details()
            # if person_features is not None:
            #     person_msg.attributes = person_features

            out_msg.people.append(person_msg) 

        return out_msg
    

if __name__ == "__main__":
    try:
        rospy.init_node('unsw_person_recognition_node')
        pr = PersonRecogniser()        
        rospy.spin()
    except (KeyboardInterrupt, rospy.ROSInterruptException) as e:
            rospy.logfatal("Error occurred! Stopping the unsw person recognition node...")
            rospy.sleep(1)
            print("unsw_person_recognition_node terminated")
   
