#!/usr/bin/env python3
import rospy
import math
from datetime import datetime
import tf.transformations
import tf2_ros
import tf

# msg imports 
from nav_msgs.msg import Odometry
import tf2_geometry_msgs
from sensor_msgs.msg import JointState
from tmc_msgs.msg import BatteryState
from tmc_control_msgs.msg import ServoState
from geometry_msgs.msg import WrenchStamped

# WM imports 
import os
import sys
import json
from dbase import DBase
from robot_state_api import RobotStateApi
module_path = os.environ.get("UNSW_WS")
sys.path.append(module_path + "/WORLD_MODEL/world_model/world_model_kb/scripts/" + "database.ini")

class RobotState:
    def __init__(self):
        self.frame_id = 1

        self.base_pose_x = 'null'
        self.base_pose_y = 'null'
        self.base_pose_z = 'null'
        self.region_id = 'null'
        self.base_r = 'null'
        self.base_p = 'null'
        self.base_yw = 'null'

        self.head_pose_z = 'null'
        self.head_r = 'null'
        self.head_p = 'null'
        self.head_yw = 'null'

        # end_effector pose in map frame
        self.end_effector_x = 'null'
        self.end_effector_y = 'null'
        self.end_effector_z = 'null'
        self.end_effector_r = 'null'
        self.end_effector_p = 'null'
        self.end_effector_yw = 'null'

        self.driveable_state = 'null'
        self.is_wrist_pressed = 'null'
        self.gripper_closed = 'null' # True for closed, False for open
        self.listening_state = 'null'
        self.speaking_state = 'null'
        self.battery_status = 'null' # battery status ranges from 0-100 representing a percentage charge. 

        self.holding_object_id = 'null'

class RobotStateUpdater:
    def __init__(self, tfBuffer:tf2_ros.Buffer):
        # initialise RobotState class
        self.robot = RobotState()

        # set tf_buffer
        self.tf_buffer = tfBuffer

        # initialise database connection
        self.robot_state_api = RobotStateApi()
        module_path = os.environ.get("UNSW_WS")
        self.db = DBase(module_path + "/WORLD_MODEL/world_model/world_model_kb/scripts/database.ini")  # database connections

        # Initialise subscribers
        odom_sub = rospy.Subscriber("/hsrb/odom", Odometry, self.odom_callback, queue_size=1)
        battery_sub = rospy.Subscriber("/hsrb/battery_state", BatteryState, self.battery_callback, queue_size=1)
        joint_sub = rospy.Subscriber("/hsrb/joint_states", JointState, self.joint_callback, queue_size=1)
        wrist_wrench_sub = rospy.Subscriber("/hsrb/wrist_wrench/raw",WrenchStamped, self.wrist_wrench_callback, queue_size=1)
        self.wrist_press_updated = 0

        print("initialising")

    # Callbacks
    def odom_callback(self, msg:Odometry):
        rate = rospy.Rate(10.0)

        # Construct pose stamped in odom frame for odometry message which corresponds to robots location
        odom_pose_stamped = tf2_geometry_msgs.PoseStamped()
        odom_pose_stamped.header.frame_id = msg.header.frame_id
        odom_pose_stamped.header.stamp = rospy.Time.now()
        odom_pose_stamped.pose = msg.pose.pose

        try:
            # transform from odom to map frame
            transform = self.tf_buffer.lookup_transform("odom", "map",rospy.Time(0), rospy.Duration(1.0))
            map_pose_stamped = tf2_geometry_msgs.do_transform_pose(odom_pose_stamped, transform)

        except(tf2_ros.LookupException, tf2_ros.ConnectivityException, tf2_ros.ExtrapolationException):
            raise

        self.robot.base_pose_x = map_pose_stamped.pose.position.x
        self.robot.base_pose_y = map_pose_stamped.pose.position.y
        self.robot.base_pose_z = map_pose_stamped.pose.position.z

        (self.robot.base_r, self.robot.base_p, self.robot.base_yw) = tf.transformations.euler_from_quaternion([map_pose_stamped.pose.orientation.x, map_pose_stamped.pose.orientation.y, map_pose_stamped.pose.orientation.z, map_pose_stamped.pose.orientation.w])

    def battery_callback(self, msg:BatteryState):
        self.robot.battery_status = msg.power

    def joint_callback(self, msg:JointState):
        # gripper is hand_motor joint, open position is ~1.24, closed position is -0.88
        gripper_index = msg.name.index("hand_motor_joint")
        if (msg.position[gripper_index] < 1.2):
            self.robot.gripper_closed = True
        else:
            self.robot.gripper_closed = False

        # arm_flex_joint -0.84
        arm_lift_index = msg.name.index("arm_lift_joint")
        arm_flex_index = msg.name.index("arm_flex_joint")
        if (msg.position[arm_lift_index] > 0.3 or msg.position[arm_flex_index] < -0.84):
            self.robot.driveable_state = False
        else:
            self.robot.driveable_state = True

    def wrist_wrench_callback(self, msg:WrenchStamped):
        if msg.wrench.force.x > 20:
            self.robot.is_wrist_pressed = True
            self.wrist_press_updated = msg.header.stamp.secs
        elif (msg.header.stamp.secs - self.wrist_press_updated) > 10 and msg.wrench.force.x < 9:
            self.robot.is_wrist_pressed = False
            

    def head_updater(self):
        try:
            head_transform = self.tf_buffer.lookup_transform("map", "head_center_camera_frame", rospy.Time(0), rospy.Duration(1.0))
            head_pos = head_transform.transform.translation
            head_rot = head_transform.transform.rotation

            self.robot.head_pose_z = head_pos.z
            [self.robot.head_r, self.robot.head_p, self.robot.head_yw] = tf.transformations.euler_from_quaternion([head_rot.x, head_rot.y, head_rot.z, head_rot.w])
        except(tf2_ros.LookupException, tf2_ros.ConnectivityException, tf2_ros.ExtrapolationException):
            rospy.logerr(f"Could not find map->head transformation")
            raise
    
    def end_effector_updater(self):
        try:
            # TODO could change to hand_l_finger_tip_frame or hand_r_finger_tip_frame, etc
            end_effector_transform = self.tf_buffer.lookup_transform("map", "hand_palm_link", rospy.Time(0), rospy.Duration(1.0))
            end_effector_pos = end_effector_transform.transform.translation
            end_effector_rot = end_effector_transform.transform.rotation

            self.robot.end_effector_x = end_effector_pos.x
            self.robot.end_effector_y = end_effector_pos.y
            self.robot.end_effector_z = end_effector_pos.z
            [self.robot.end_effector_r, self.robot.end_effector_p, self.robot.end_effector_yw] = tf.transformations.euler_from_quaternion([end_effector_rot.x, end_effector_rot.y, end_effector_rot.z, end_effector_rot.w])
        except(tf2_ros.LookupException, tf2_ros.ConnectivityException, tf2_ros.ExtrapolationException):
            rospy.logerr(f"Could not find map->hand transformation")
            raise
            

    def update_wm(self):
        self.head_updater()
        self.end_effector_updater()
        robot_state_dict = {
            "frame_id" : self.robot.frame_id,

            "base_pose_x" : self.robot.base_pose_x,
            "base_pose_y" : self.robot.base_pose_y, 
            "base_pose_z" : self.robot.base_pose_z, 
            "region_id" : self.robot.region_id,
            "base_r" : self.robot.base_r, 
            "base_p" : self.robot.base_p, 
            "base_yw" : self.robot.base_yw,

            "head_pose_z" : self.robot.head_pose_z,
            "head_r" : self.robot.head_r,
            "head_p" : self.robot.head_p,
            "head_yw" : self.robot.head_yw,

            "end_effector_x" : self.robot.end_effector_x,
            "end_effector_y" : self.robot.end_effector_y,
            "end_effector_z" : self.robot.end_effector_z,
            "end_effector_r" : self.robot.end_effector_r,
            "end_effector_p" : self.robot.end_effector_p,
            "end_effector_yw" : self.robot.end_effector_yw,

            "driveable_state" : self.robot.driveable_state, 
            "is_wrist_pressed" : self.robot.is_wrist_pressed,
            "gripper_closed" : self.robot.gripper_closed,
            "listening_state" : self.robot.listening_state,
            "speaking_state" : self.robot.speaking_state,
            "battery_status" : self.robot.battery_status,

            "holding_object_id" : self.robot.holding_object_id,
        }   

        self.robot_state_api.insert_robot_state(self.db.con_pool, json.dumps(robot_state_dict))
        rospy.loginfo("Robot state updated")

def main():
    # initialise node
    rospy.init_node("robot_state_updater")
    
    tfBuffer = tf2_ros.Buffer()
    listener = tf2_ros.TransformListener(tfBuffer)
    
    robot_state_updater = RobotStateUpdater(tfBuffer)
    while not rospy.is_shutdown():
        robot_state_updater.update_wm()
        rospy.sleep(0.2)

if __name__ == '__main__':
    main()


